import math, random, os, sys, argparse, csv, logging, shutil
from datetime import datetime
import numpy as np
import scipy.stats as st
import pandas as pd

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.autograd as autograd
import torchvision
from torchvision import datasets, models, transforms
from PIL import Image

# from tqdm import tqdm, tqdm_notebook
from delete_VNRF import VNRF_rmpercent, get_accuracy


def parse_arguments():
    parser = argparse.ArgumentParser(description='Trains a UAP')
    parser.add_argument('--source-model', nargs="+", default=['resnet50'])
    parser.add_argument('--source-epsilon', type=str, default=0)
    parser.add_argument("--target-model", nargs="+", default=['inception_v3', 'resnet50', 'densenet121', 'vgg16_bn'])
    parser.add_argument('--batch-size', type=int, default=32, help='Batch size (default: 32)')
    parser.add_argument('--sampling-frequency', type=int, default=1, help='sampling the images to save the exploration time')
    parser.add_argument('--checkpoint', default=None, help='Path to the weight (default: None)')
    # Attack specifics
    parser.add_argument('--max-iterations', type=int, default=20, help='Maximum Iterations (default: 20)')
    parser.add_argument('--lr', type=eval, default=2./255., help='Learning rate (default: 2/255)')
    parser.add_argument('--epsilon', type=int, default=16, help='Epsilon (default: 16)')

    # VNRF
    parser.add_argument('--vnrf_layer_name', type=str)
    parser.add_argument('--vnrf_num', type=int, default=4, help='VNRF_num')

    parser.add_argument('--num_batches', type=int, default=4, help='num_batches')


    args = parser.parse_args()

    assert args.max_iterations % args.log_interval == 0
    return args

def main():
    
    args = parse_arguments()
    
    result_path = './results/{}/{}{}'.format(args.subfolder, get_timestamp(), args.postfix)
    os.makedirs(result_path)

    logger = get_logger(result_path)

    state = {k: v for k, v in args._get_kwargs()}
    for key, value in state.items():
        logger.info('{} : {}'.format(key, value))

    # Saving this file
    shutil.copy(sys.argv[0], os.path.join(result_path, sys.argv[0]))
    shutil.copytree('networks', os.path.join(result_path, 'networks'))

    # Pandas dataframe to store the results 
    column_names = ["source_model", "target_model", "iteration", "targeted", "accuracy"]
    df = pd.DataFrame(columns = column_names)

    # values are standard normalization for ImageNet images, 
    # from https://github.com/pytorch/examples/blob/master/imagenet/main.py
    norm = Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
    num_classes = 1000

    trn = transforms.Compose([
    #      transforms.Resize((256, 256)),
        transforms.ToTensor(),])

    device  = torch.device("cuda:0")

    # Source models
    source_model_names = args.source_model
    num_source_models = len(source_model_names)
    
    source_models = []
    

    for model_name in source_model_names:

        if model_name == "resnet18":
            source_model = models.resnet18(pretrained=True).eval().to(device)
            state_dict_ = VNRF_rmpercent(source_model, int(args.vnrf_layer_name))
            source_model.load_state_dict(state_dict_)
    

        for param in source_model.parameters():
            param.requires_grad=False

        source_model.to(device)
        source_models.append(source_model)

    # Target models
    # target_model_names = ['inception_v3', 'resnet50', 'densenet121', 'vgg16_bn']
    # target_model_names = ['vgg16_bn']
    target_model_names = args.target_model
    num_target_models = len(target_model_names)

    target_models = []
    for model_name in target_model_names:
        target_model = models.__dict__[model_name](pretrained=True).eval()

        for param in target_model.parameters():
            param.requires_grad=False

        target_model.to(device)
        target_models.append(target_model)

    seed_num=1
    random.seed(seed_num)
    np.random.seed(seed_num)
    torch.manual_seed(seed_num)
    torch.backends.cudnn.deterministic = True

    # Transformation results

    clean_logits_list = []
    adv_logits_list = []
    for _ in range(num_target_models):
        clean_logits = torch.zeros(num_samples, 1000)
        adv_logits = torch.zeros(num_samples, 1000)

        clean_logits_list.append(clean_logits)
        adv_logits_list.append(adv_logits)

    if args.store_samples:
        adv_examples_storage = torch.zeros(num_samples, 3, args.img_size, args.img_size)

    # num_batches = args.num_batches
    for k in range(0,num_batches):
        print(k, num_batches)
        batch_size_cur=min(args.batch_size,len(image_id_list)-k*args.batch_size)        
        X_ori = torch.zeros(batch_size_cur,3,args.img_size,args.img_size).to(device)
        delta= torch.zeros_like(X_ori,requires_grad=True).to(device)

        labels_tar=torch.tensor(label_tar_list[k*args.batch_size*sampling_frequency:(k*args.batch_size+batch_size_cur)*sampling_frequency:sampling_frequency]).to(device)
        labels_gt=torch.tensor(label_ori_list[k*args.batch_size*sampling_frequency:(k*args.batch_size+batch_size_cur)*sampling_frequency:sampling_frequency]).to(device)
        
        grad_momentum=0
        prev = float('inf')
        for t in range(args.max_iterations):
            if args.di:
                # import pdb; pdb.set_trace()
                X_adv = DI(X_ori+delta)
            else:
                X_adv = X_ori+delta

            if args.input_noise == 'uniform':
                # X_adv = weight * X_adv
                pert_noise = torch.zeros_like(X_adv).uniform_(-args.input_noise_std, args.input_noise_std)
                X_adv = torch.clamp(X_adv + pert_noise, 0, 1)

            logits = 0
            for source_model in source_models:
                logits += source_model(norm(X_adv))
            logits /= num_source_models

            if t % args.log_interval == (args.log_interval - 1):
                pos_ens_ori[0, t // args.log_interval]=pos_ens_ori[0, t // args.log_interval]+torch.sum(torch.argmax(logits,dim=1)==labels_gt).cpu().numpy()
                pos_ens_tar[0, t // args.log_interval]=pos_ens_tar[0, t // args.log_interval]+torch.sum(torch.argmax(logits,dim=1)==labels_tar).cpu().numpy()
            
            
            loss = -nn.CrossEntropyLoss()(logits,labels_gt)
            loss.backward()


            grad_a = delta.grad.clone()
            delta.grad.zero_()
            
            delta.data=((X_ori+delta.data).clamp(0,1))-X_ori


                for model_idx in range(num_target_models):
                    pos_ori[model_idx, t // args.log_interval]=pos_ori[model_idx, t // args.log_interval]+torch.sum(torch.argmax(target_models[model_idx](norm(X_ori+delta)),dim=1)==labels_gt).cpu().numpy()
                    pos_tar[model_idx, t // args.log_interval]=pos_tar[model_idx, t // args.log_interval]+torch.sum(torch.argmax(target_models[model_idx](norm(X_ori+delta)),dim=1)==labels_tar).cpu().numpy()

                    target_model = target_models[model_idx]
                    normal_out = target_model(norm(X_ori))
                    adv_out = target_model(norm(X_ori+delta))

                    normal_out = normal_out.cpu()
                    adv_out = adv_out.cpu()

                    normal_top = torch.argmax(normal_out, dim=-1)
                    adv_top = torch.argmax(adv_out, dim=-1)
        number_seen_samples = (k+1) * args.batch_size
        logger.info('\n-- (Batch {}/{}) (Iteration {}/{}) Untargeted Accuracy --'.format(k+1, num_batches, t+1, args.max_iterations))
        for i in range(num_target_models):
            logger.info('{} -> {}'.format(args.source_model, target_model_names[i]))
            logger.info('{}'.format(1.-pos_ori[i,:]/number_seen_samples))


        logger.info('\n--  (Batch {}/{}) (Iteration {}/{}) Targeted Accuracy --'.format(k+1, num_batches, t+1, args.max_iterations))
        for i in range(num_target_models):
            logger.info('{} -> {}'.format(args.source_model, target_model_names[i]))
            logger.info('{}'.format(pos_tar[i,:]/number_seen_samples))

        logger.info('\n-- (Batch {}/{}) (Iteration {}/{}) Untargeted Accuracy (white-box ensemble)--'.format(k+1, num_batches, t+1, args.max_iterations))
        logger.info('{}'.format(1.-pos_ens_ori[0,:]/number_seen_samples))


        logger.info('\n--  (Batch {}/{}) (Iteration {}/{}) Targeted Accuracy (white-box ensemble)--'.format(k+1, num_batches, t+1, args.max_iterations))
        logger.info('{}'.format(pos_ens_tar[0,:]/number_seen_samples))


    # Results
    logger.info('\n-- Untargeted Accuracy --')
    for i in range(num_target_models):
        logger.info('{} -> {}'.format(args.source_model, target_model_names[i]))
        logger.info('{}'.format(1.-pos_ori[i,:]/num_samples))
        

    logger.info('\n-- Targeted Accuracy --')
    for i in range(num_target_models):
        logger.info('{} -> {}'.format(args.source_model, target_model_names[i]))
        logger.info('{}'.format(pos_tar[i,:]/num_samples))

if __name__ == '__main__':
    main()